@@ -1,6 +1,7 @@
 from collections import defaultdict, Hashable
 from datetime import datetime
-from pandas import Categorical, Series
+from pandas import Categorical, DataFrame, Series
+from scipy.interpolate import BSpline
 from scipy.sparse import lil_matrix
 from sklearn.base import BaseEstimator, TransformerMixin
 from sklearn.pipeline import Pipeline
@@ -59,6 +60,21 @@ class Aggregator(BaseEstimator, TransformerMixin):
 			raise ValueError(self.function)
 		return _col2d(Xt)
 
+class BSplineTransformer(BaseEstimator, TransformerMixin):
+
+	def __init__(self, bspline):
+		if not isinstance(bspline, BSpline):
+			raise ValueError("The spline object is not an instance of {0}".format(BSpline.__name__))
+		self.bspline = bspline
+
+	def fit(self, X, y = None):
+		X = ensure_1d(X)
+		return self
+
+	def transform(self, X):
+		X = ensure_1d(X)
+		return self.bspline(X)
+
 class CastTransformer(BaseEstimator, TransformerMixin):
 	"""Change data type."""
 
@@ -95,6 +111,18 @@ class CutTransformer(BaseEstimator, TransformerMixin):
 			Xt = numpy.asarray(Xt)
 		return _col2d(Xt)
 
+class DataFrameConstructor(BaseEstimator, TransformerMixin):
+
+	def __init__(self, columns, dtype):
+		self.columns = columns
+		self.dtype = dtype
+
+	def fit(self, X, y = None):
+		return self
+
+	def transform(self, X):
+		return DataFrame(X, columns = self.columns, dtype = self.dtype)
+
 class DurationTransformer(BaseEstimator, TransformerMixin):
 	"""Calculate time difference."""
 
@@ -154,21 +182,56 @@ class SecondsSinceMidnightTransformer(BaseEstimator, TransformerMixin):
 		return dt_transform(X, to_int_duration)
 
 class ExpressionTransformer(BaseEstimator, TransformerMixin):
-	"""Transform data using a Python expression."""
+	"""Transform data using a Python expression.
 
-	def __init__(self, expr, dtype = None):
+	Parameters:
+	----------
+	map_missing_to: scalar, optional
+		The return value when any of the expression arguments is missing.
+
+	default_value: scalar, optional
+		The return value when the expression result is missing.
+
+	invalid_value_treatment: string
+		The action to take when the evaluation of the expression raises an error.
+	"""
+
+	def __init__(self, expr, map_missing_to = None, default_value = None, invalid_value_treatment = None, dtype = None):
 		self.expr = expr
+		self.map_missing_to = map_missing_to
+		self.default_value = default_value
+		invalid_value_treatments = ["return_invalid", "as_missing"]
+		if (invalid_value_treatment is not None) and (invalid_value_treatment not in invalid_value_treatments):
+			raise ValueError("Invalid value treatment {0} not in {1}".format(invalid_value_treatment, invalid_value_treatments))
+		self.invalid_value_treatment = invalid_value_treatment
 		self.dtype = dtype
 
 	def _eval_row(self, X):
-		return eval(self.expr)
+		# X is array-like (row vector)
+		if (self.map_missing_to is not None) and ((pandas.isnull(X)).any()):
+			return self.map_missing_to
+		try:
+			Xt = eval(self.expr)
+		except ArithmeticError as ae:
+			if self.invalid_value_treatment == "return_invalid":
+				raise ae
+			elif self.invalid_value_treatment == "as_missing":
+				Xt = None
+			else:
+				pass
+		# Xt is scalar
+		if (self.default_value is not None) and (pandas.isnull(Xt)):
+			return self.default_value
+		return Xt
 
 	def fit(self, X, y = None):
 		return self
 
 	def transform(self, X):
 		func = lambda x: self._eval_row(x)
-		Xt = eval_rows(X, func)
+		# Evaluate in PMML compatibility mode
+		with numpy.errstate(divide = "raise"):
+			Xt = eval_rows(X, func)
 		if self.dtype is not None:
 			Xt = cast(Xt, self.dtype)
 		return _col2d(Xt)
